from setuptools import setup, find_packages

setup(
    name="oass",
    version="1.0.0",
    author="Artiprocher",
    author_email="zjduan@stu.ecnu.edu.cn",
    description="Optimal Action Space Search (OASS) is an algorithm for path planning problems on directed acyclic graphs (DAG) based on reinforcement learning (RL) theory.",
    url="https://github.com/ECNU-CILAB/OASS", 
    packages=find_packages()
)
